#!/usr/bin/env python3

# Copyright (C) 2024 Roberto Rossini <roberros@uio.no>
#
# SPDX-License-Identifier: MIT

import argparse
import logging
import multiprocessing as mp
import pathlib
import shutil
import subprocess as sp
import tempfile
from typing import Dict, List, Union

import hictkpy as htk
from packaging.version import Version


def existing_file(arg: str) -> pathlib.Path:
    if (path := pathlib.Path(arg)).is_file():
        return path

    raise FileNotFoundError(arg)


def positive_int(arg) -> int:
    if (n := int(arg)) > 0:
        return n

    raise ValueError("Not a positive int")


def num_cpus(arg: str) -> int:
    try:
        n = int(arg)
        if 0 < n <= mp.cpu_count():
            return n
    except:  # noqa
        pass

    raise ValueError(f"Not a valid number of CPU cores (allowed values are integers between 1 and {mp.cpu_count()})")


def make_cli() -> argparse.ArgumentParser:
    cli = argparse.ArgumentParser(
        "Given a matrix in .mcool or .hic format, generate a matrix file suitable for testing."
    )

    cli.add_argument(
        "input-matrix",
        type=existing_file,
        help="Path to a Hi-C matrix in .cool, .mcool or .hic format.",
    )
    cli.add_argument(
        "output-matrix",
        type=pathlib.Path,
        help="Path where to store the resulting output matrix file.",
    )
    cli.add_argument(
        "--resolutions",
        nargs="*",
        type=positive_int,
        help="One or more resolutions to be generated.\n"
        "When more than one resolution is provided, all resolutions should be multiple of the base\n"
        "(i.e. smallest) resolution.\n"
        "At least one resolution is required when input file is not in .cool format.",
    )
    cli.add_argument(
        "--hictk-bin",
        type=existing_file,
        help="Path to the hictk binary. When not provided, hictk's binary is looked up in the system PATH.",
    )
    cli.add_argument(
        "--genomic-belt",
        type=int,
        default=5_000_000,
        help="Radius of the band, centred around the diagonal. Interactions outside of this band are ignored.",
    )
    cli.add_argument(
        "-p",
        "--nproc",
        type=num_cpus,
        default=1,
        help="Maximum number of parallel processes to use. Only used when processing .hic files.",
    )

    cli.add_argument(
        "--force",
        action="store_true",
        default=False,
        help="Overwrite existing file(s).",
    )

    return cli


def find_hictk(path: Union[pathlib.Path, None]) -> pathlib.Path:
    if path is None:
        path = "hictk"

    if shutil.which(path):
        return path

    raise FileNotFoundError("unable to locate hictk's binary")


def dump_chroms(chroms: Dict[str, int], tmpdir: pathlib.Path) -> pathlib.Path:
    path = tmpdir / "chroms.sizes"
    with open(path, "w") as f:
        for chrom, size in chroms.items():
            f.write(f"{chrom}\t{size}\n")

    return path


def run_hictk_load(
    f: htk.File, genomic_belt: int, tmpdir: pathlib.Path, threads: int, skip_all_vs_all_matrix: bool
) -> pathlib.Path:
    if f.is_hic():
        path = tmpdir / "matrix.hic"
        w = htk.hic.FileWriter(
            path,
            f.chromosomes(),
            f.resolution(),
            f.attributes()["assembly"],
            tmpdir=tmpdir,
            compression_lvl=12,
            skip_all_vs_all_matrix=skip_all_vs_all_matrix,
            n_threads=threads,
        )
    else:
        path = tmpdir / "matrix.cool"
        w = htk.cooler.FileWriter(
            path, f.chromosomes(), f.resolution(), f.attributes()["assembly"], tmpdir=tmpdir, compression_lvl=9
        )

    genomic_belt = (genomic_belt - f.resolution() - 1) // f.resolution()

    for chrom in f.chromosomes():
        logging.info("fetching interactions for %s...", chrom)
        df = f.fetch(chrom).to_pandas()
        num_interactions = len(df)
        df = df[df["bin2_id"] - df["bin1_id"] < genomic_belt]

        logging.info("dropped %d out of %d interactions", num_interactions - len(df), num_interactions)
        w.add_pixels(df)

    if Version(htk.__version__) > Version("1.0.0"):
        # Workaround data race fixed by https://github.com/paulsengroup/hictkpy/pull/131
        w.finalize(log_lvl="INFO")
    else:
        w.finalize(log_lvl="CRITICAL")

    return path


def run_hictk_zoomify(
    hictk: pathlib.Path,
    input_path: pathlib.Path,
    output_path: pathlib.Path,
    resolutions: List[int],
    tmpdir: pathlib.Path,
    threads: int,
    force: bool,
):
    assert len(resolutions) > 1

    resolutions = list(map(str, resolutions))

    cmd = [
        hictk,
        "zoomify",
        input_path,
        output_path,
        "--resolutions",
        *resolutions,
        "--tmpdir",
        tmpdir,
        "--threads",
        str(threads),
    ]
    if force:
        cmd.append("--force")

    logging.info("running %s", " ".join(str(x) for x in cmd))
    sp.check_call(cmd)


def process_resolutions(resolutions: List[int]) -> List[int]:
    resolutions = set(resolutions)
    resolutions = list(sorted(resolutions))

    for res in resolutions:
        if res % resolutions[0] != 0:
            raise RuntimeError(f"resolution {res} is not a multiple of {resolutions[0]}")

    return resolutions


def main():
    args = vars(make_cli().parse_args())

    if args["output-matrix"].exists() and not args["force"]:
        raise RuntimeError(f"refusing to overwrite file {args['output-matrix']}. Pass --force to overwrite.")

    if args["resolutions"] is None:
        if htk.is_cooler(args["input-matrix"]):
            resolutions = [htk.File(args["input-matrix"]).resolution()]
        else:
            raise RuntimeError("--resolutions is required when input matrix is not in .cool format.")
    else:
        resolutions = process_resolutions(args["resolutions"])

    need_to_zoomify = len(resolutions) > 1

    if need_to_zoomify:
        hictk = find_hictk(args["hictk_bin"])

    f = htk.File(args["input-matrix"], resolution=resolutions[0])
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir = pathlib.Path(tmpdir)
        single_res_file = run_hictk_load(
            f, args["genomic_belt"], tmpdir, args["nproc"], skip_all_vs_all_matrix=need_to_zoomify
        )

        if not need_to_zoomify:
            single_res_file.rename(args["output-matrix"])
            return

        run_hictk_zoomify(
            hictk,
            single_res_file,
            args["output-matrix"],
            resolutions,
            tmpdir,
            args["nproc"],
            args["force"],
        )


def setup_logger(level: str):
    fmt = "[%(asctime)s] %(levelname)s: %(message)s"
    logging.basicConfig(level=level, format=fmt)
    logging.getLogger().setLevel(level)


if __name__ == "__main__":
    setup_logger("INFO")

    main()



# Copyright (C) 2024 Andrea Raffo <andrea.raffo@ibv.uio.no>
#
# SPDX-License-Identifier: MIT

# Implementation from the library by Tino Weinkauf downloadable at:
# https://www.csc.kth.se/~weinkauf/notes/persistence1d.html


class UnionFind:
    """
    Implements the Union-Find data structure.

    It keeps track of elements partitioned into a number of disjoint (non-overlapping) sets.
    This data structure is required for merge trees and similar algorithms.

    This implementation uses path compression in several places, written with a merge-tree-like algorithm in mind.

    A set is identified by the ID of the defining element (e.g., vertex).

    Author: Tino Weinkauf
    """

    NOSET = -1

    def __init__(self, NumElements):
        """Initializes the domain with @NumElements elements living in zero sets."""
        self.Domain = [int(UnionFind.NOSET)] * NumElements
        self.NumSets = 0

    def _assert(self, condition, text):
        if not condition:
            raise ValueError(text)

    def GetNumSets(self):
        """Returns the number of sets."""
        return self.NumSets

    def MakeSet(self, idxElement):
        """Creates a new set with the given @idxElement as a root."""
        self._assert(self.Domain[idxElement] == UnionFind.NOSET, "You cannot call MakeSet() on an existing set.")

        self.Domain[idxElement] = idxElement
        self.NumSets += 1

    def ExtendSet(self, idxElementFrom, idxElementTo):
        """
        Extends a set from one element to the next.

        @note The element identified by @idxElementFrom needs to belong to a set.
        @note The element identified by @idxElementTo cannot belong to a set.
        """
        self._assert(self.Domain[idxElementTo] == UnionFind.NOSET, "You cannot call ExtendSet() on an existing set.")
        self._assert(
            self.Domain[idxElementFrom] != UnionFind.NOSET, "You cannot call ExtendSet() without an existing set."
        )

        idxRoot = self.Find(idxElementFrom)
        self.Domain[idxElementTo] = idxRoot
        self.Domain[idxElementFrom] = idxRoot  # compression only; not functionally required.

    def ExtendSetByID(self, idxRoot, idxElementTo):
        """
        Extends a set with a given set ID.

        @note The set identified by @idxRoot needs to exist.
            In particular, this needs to be true: Find(idxRoot) == idxRoot

        @note The element identified by @idxElementTo cannot belong to a set.
        """
        self._assert(
            self.Domain[idxElementTo] == UnionFind.NOSET, "You cannot call ExtendSetByID() on an existing set."
        )
        self._assert(
            self.Domain[idxRoot] != UnionFind.NOSET, "You cannot call ExtendSetByID() without an existing set."
        )
        self._assert(idxRoot == self.Find(idxRoot), "Please call the function ExtendSetByID() with an existing set ID.")

        self.Domain[idxElementTo] = idxRoot

    def Union(self, idxElementMergeThisOne, idxElementIntoThisOne):
        """
        Merges two sets into one.

        The two sets are identified by their elements @idxElementMergeThisOne and @idxElementIntoThisOne.
        The former set is merged into the latter, i.e., the latter one remains.

        This function uses a lot of compression to speed-up later calls to any Find*() function.
        """
        self._assert(
            idxElementMergeThisOne != UnionFind.NOSET and idxElementIntoThisOne != UnionFind.NOSET,
            "You can only call Union() on existing sets.",
        )

        # ~ Find the root that will survive this union
        SetIDB = self.FindAndCompress(idxElementIntoThisOne)

        # ~ Find the root that will merge into the surviving one, compress/merge along the way
        bIsSameSet = self.FindMergeCompressTo(idxElementMergeThisOne, SetIDB)

        # ~ Same set?
        if not bIsSameSet:
            self.NumSets -= 1

        self._assert(self.NumSets > 0, "We should have at least one set.")

    def Find(self, idxElement):
        """
        Finds the ID of the set to which the element @idxElement belongs.

        This function does not use compression, and therefore does not change any underlying data.
        """
        if self.Domain[idxElement] == UnionFind.NOSET:
            return UnionFind.NOSET

        # ~ Find the root of the set
        while True:
            idxElement = self.Domain[idxElement]
            if idxElement == self.Domain[idxElement]:
                break

        # ~ Return root of set == ID of set
        return idxElement

    def FindAndCompress(self, idxElement):
        """
        Finds the ID of the set to which the element @idxElement belongs, and compresses the entire path.

        Compression means that all elements along the path point to the root of the set.
        This makes future calls to any Find*() function faster.
        """
        if self.Domain[idxElement] == UnionFind.NOSET:
            return UnionFind.NOSET

        # ~ Record a path
        Path = [idxElement]

        # ~ Find the root of the set
        while True:
            idxElement = self.Domain[idxElement]
            Path.append(idxElement)
            if idxElement == self.Domain[idxElement]:
                break

        # ~ Compress the path
        for idxStep in Path:
            self.Domain[idxStep] = idxElement

        # ~ Return root of set == ID of set
        return idxElement

    def FindMergeCompressTo(self, idxElement, idxRoot):
        """
        Find a path from @idxElement to its root and compresses the entire path to a new root.

        Useful only when merging sets.

        @returns true, if the root of @idxElement is already idxRoot, i.e., they belong to the same set.
        @returns false, otherwise.
        """
        if self.Domain[idxElement] == UnionFind.NOSET:
            return False

        # ~ Record a path
        Path = [idxElement]

        # ~ Find the root of the set
        while True:
            idxElement = self.Domain[idxElement]
            Path.append(idxElement)
            if idxElement == self.Domain[idxElement]:
                break

        bIsSameSet = idxElement == idxRoot

        # ~ Compress the path
        for idxStep in Path:
            self.Domain[idxStep] = idxRoot

        return bIsSameSet



# Copyright (C) 2024 Andrea Raffo <andrea.raffo@ibv.uio.no>
#
# SPDX-License-Identifier: MIT

import numpy as np

from .unionfind import UnionFind

# Implementation adapted from the library by Tino Weinkauf downloadable at:
# https://www.csc.kth.se/~weinkauf/notes/persistence1d.html

# ATT: in its original implementation, the terms minima/maxima/extrema were sometimes used in place of
# minimum/maximum/extremum points. This notation is here fixed.
# For more details, consult: https://en.wikipedia.org/wiki/Maximum_and_minimum


def run_persistence(data, level_sets="lower"):
    """
    This function finds local extrema and their persistence in one-dimensional data w.r.t. lower (default,
    level_sets="lower") or upper (level_sets="upper") level sets.

    Local minima and local maxima are extracted, paired, and returned together with their persistence.
    For level_sets="lower", the global minimum is extracted as well.
    For level_sets="upper", the global maximum is extracted as well.

    We assume a connected one-dimensional domain.

    Short explanation for the case of level_sets=="lower" (the case of level_sets="upper" is analogous).

    This function returns a list of extrema together with their persistence. The list is NOT sorted, but the paired
    extrema can be identified, i.e., which minimum and maximum were removed together at a particular persistence
    level. As follows:
    (*)  Odd entries are minima, even entries are maxima.
    (*)  The minimum at 2*i is paired with the maximum at 2*i+1.
    (*)  The last entry of the list is the global minimum (resp. maximum) when level_sets="lower" (resp.
         level_sets="upper"). It is not paired with a maximum (resp. minimum).
    Hence, the list has an odd number of entries.

    Authors: Tino Weinkauf (original implementation) and Andrea Raffo (modified implementation)
    """

    # Number of data to break ties (leftmost index comes first):
    num_elements = len(data)
    sorted_idx = np.argsort(data, kind="stable")[::-1] if level_sets == "upper" else np.argsort(data, kind="stable")

    # Get a union find data structure:
    uf = UnionFind(num_elements)

    # Extrema paired with topological persistence:
    extremum_points_and_persistence = []

    # Watershed:
    for idx in sorted_idx:

        # Get neighborhood indices:
        left_idx = max(idx - 1, 0)
        right_idx = min(idx + 1, num_elements - 1)

        # Count number of components in neighborhood:
        neighbor_components = [
            uf.Find(neighbor) for neighbor in [left_idx, right_idx] if uf.Find(neighbor) != UnionFind.NOSET
        ]
        num_neighbor_components = len(neighbor_components)

        if num_neighbor_components == 0:
            # Create a new component:
            uf.MakeSet(idx)
        elif num_neighbor_components == 1:
            # Extend the one and only component in the neighborhood
            # Note that NeighborComponents[0] holds the root of a component, since we called Find() earlier to retrieve
            # it!
            uf.ExtendSetByID(neighbor_components[0], idx)
        else:

            if level_sets == "lower":
                # Merge the two components on either side of the current point:
                idx_lowest_minimum = neighbor_components[np.argmin(data[neighbor_components])]
                idx_highest_minimum = [comp for comp in neighbor_components if comp != idx_lowest_minimum][0]
                uf.ExtendSetByID(idx_lowest_minimum, idx)
                uf.Union(idx_highest_minimum, idx_lowest_minimum)

                # Record the two paired extrema: index of minimum, index of maximum, persistence value:
                persistence = data[idx] - data[idx_highest_minimum]
                extremum_points_and_persistence.append((idx_highest_minimum, persistence))
                extremum_points_and_persistence.append((idx, persistence))

            elif level_sets == "upper":

                # Merge the two components on either side of the current point:
                idx_highest_maximum = neighbor_components[np.argmax(data[neighbor_components])]
                idx_lowest_maximum = [comp for comp in neighbor_components if comp != idx_highest_maximum][0]
                uf.ExtendSetByID(idx_highest_maximum, idx)
                uf.Union(idx_lowest_maximum, idx_highest_maximum)

                # Record the two paired extrema: index of minimum, index of maximum, persistence value:
                persistence = data[idx_lowest_maximum] - data[idx]
                extremum_points_and_persistence.append((idx, persistence))
                extremum_points_and_persistence.append((idx_lowest_maximum, persistence))

    # Global minimum (or maximum):
    if level_sets == "lower":
        extremum_points_and_persistence.append((uf.Find(0), np.inf))
    elif level_sets == "upper":
        extremum_points_and_persistence.append((uf.Find(0), np.inf))

    return extremum_points_and_persistence


def diversify_extremum_points_and_persistence(extremum_points_and_persistence, level_set):
    min_points_and_persistence = [t for t in extremum_points_and_persistence[::2]]
    max_points_and_persistence = [t for t in extremum_points_and_persistence[1::2]]

    if level_set == "upper":
        max_points_and_persistence = max_points_and_persistence + [min_points_and_persistence[-1]]
        min_points_and_persistence = min_points_and_persistence[:-1]

    return min_points_and_persistence, max_points_and_persistence


def filter_extremum_points_by_persistence(extremum_points_and_persistence, threshold):

    filtered_extremum_points_and_persistence = [
        (point, persistence) for point, persistence in extremum_points_and_persistence if persistence > threshold
    ]
    return filtered_extremum_points_and_persistence



# Copyright (C) 2024 Roberto Rossini <roberroso@uio.no>
#
# SPDX-License-Identifier: MIT

import logging
import sys
from typing import List, Union

from .cli import call, download, plot, setup, view


def _setup_mpl_backend():
    # This is very important, as some plotting operations are performed concurrently
    # using multiprocessing.
    # If the wrong backend is selected (e.g. tkinter) this can lead to the whole OS freezing
    import matplotlib

    matplotlib.use("Agg")


def _setup_logger(level: str):
    fmt = "[%(asctime)s] %(levelname)s: %(message)s"
    logging.basicConfig(level=level, format=fmt)
    logging.getLogger().setLevel(level)


def main(args: Union[List[str], None] = None):
    if args is None:
        args = sys.argv[1:]

    subcommand, args = setup.parse_args(args)
    _setup_logger("INFO")  # TODO make tunable

    if subcommand == "call":
        _setup_mpl_backend()
        return call.run(**args)
    if subcommand == "download":
        return download.run(**args)
    if subcommand == "plot":
        _setup_mpl_backend()
        return plot.run(**args)
    if subcommand == "view":
        return view.run(**args)

    raise NotImplementedError


if __name__ == "__main__":
    main()



# Copyright (C) 2024 Roberto Rossini <roberroso@uio.no>
#
# SPDX-License-Identifier: MIT

import math
from typing import Tuple, Union

import numpy as np
import scipy.sparse as ss
from numpy.typing import NDArray


class Stripe(object):
    """
    A class used to represent architectural stripes.
    This class takes care of validating stripe coordinates and computing several descriptive statistics.

    This is how this class should be used:

        * Initialize the class by providing at least the seed position
        * At a later time, set the vertical and horizontal boundaries by calling `set_horizontal_bounds()` and `set_vertical_bounds()`
        * Finally, call `compute_biodescriptors()` to compute and store the descriptive statistics

    The stripe properties and statistics can now be accessed through the attributes listed below.

    Attributes representing the descriptive statistics return negative values to signal that it was not
    possible to compute the statistics for the current Stripe instance.

    Attributes
    ----------

    seed: int
        the stripe seed
    top_persistence: Union[float, None]
        the topological persistence
    lower_triangular: bool
        true when the stripe extends in the lower-triangular portion of the matrix
    upper_triangular: bool
        true when the stripe extends in the upper-triangular portion of the matrix
    left_bound: int
        the left bound of the stripe
    right_bound: int
        the right bound of the stripe
    top_bound: int
        the top bound of the stripe
    bottom_bound: int
        the bottom bound of the stripe
    inner_mean: float
        the average number of interactions within the stripe
    inner_std: float
        the standard deviation of the number of interactions within the stripe
    five_number: NDArray[float]
        a vector of five numbers corresponding to the 0, 25, 50, 75, and 100 percentiles of the number of within-stripe interactions
    outer_lmean: float
        the average number of interactions in the band to the left of the stripe
    outer_rmean: float
        the average number of interactions in the band to the right of the stripe
    outer_mean: float
        the average number of interactions in the bands to the left and right of the stripe
    rel_change: float
        the ratio of the average number of interactions within the stripe and in the neighborhood outside of the stripe
    """

    def __init__(
        self,
        seed: int,
        top_pers: Union[float, None],
        horizontal_bounds: Union[Tuple[int, int], None] = None,
        vertical_bounds: Union[Tuple[int, int], None] = None,
        where: Union[str, None] = None,
    ):
        """
        Parameters
        ----------
        seed: int
            the stripe seed position
        top_pers: Union[float, None]
            the topological persistence of the seed
        horizontal_bounds: Union[Tuple[int, int], None]
            the horizontal bounds of the stripe
        vertical_bounds: Union[Tuple[int, int], None]
            the_vertical bounds of the stripe
        where: Union[str, None]
            the location of the stripe: should be "upper_triangular" or "lower_triangular".
            When provided, this is validate the coordinates set when calling `set_horizontal_bounds()` and `set_vertical_bounds()`.
        """
        if seed < 0:
            raise ValueError("seed must be a non-negative integral number")
        self._seed = seed

        if top_pers is not None and top_pers < 0:
            raise ValueError("when not None, top_pers must be a positive number")
        self._persistence = top_pers

        valid_locations = {"upper_triangular", "lower_triangular"}
        if where is not None and where not in valid_locations:
            raise ValueError(f"when specified, where must be one of {tuple(valid_locations)}")
        self._where = where

        self._left_bound = None
        self._right_bound = None
        if horizontal_bounds is not None:
            self.set_horizontal_bounds(*horizontal_bounds)

        self._bottom_bound = None
        self._top_bound = None
        if vertical_bounds is not None:
            self.set_vertical_bounds(*vertical_bounds)

        self._five_number = None
        self._inner_mean = None
        self._inner_std = None
        self._outer_lmean = None
        self._outer_rmean = None

    @staticmethod
    def _infer_location(seed: int, top_bound: int, bottom_bound: int) -> str:
        # TODO this check is temporarily disabled as it fails when processing stripes from chromosomes that are mostly empty
        # if bottom_bound == top_bound:
        #    raise ValueError(f"unable to infer stripe location: stripe bottom and top bounds are identical ({top_bound})")

        if bottom_bound > seed:
            return "lower_triangular"
        # TODO the equal check should be removed as is not correct
        if top_bound <= seed:
            return "upper_triangular"

        NotImplementedError

    def _compute_convex_comp(self) -> int:
        cfx1 = 0.99
        cfx2 = 0.01

        if self.upper_triangular:
            cfx1, cfx2 = cfx2, cfx1

        return int(round(cfx1 * self._top_bound + cfx2 * self._bottom_bound))

    def _slice_matrix(self, I: ss.csr_matrix) -> NDArray:
        convex_comb = self._compute_convex_comp()

        if self.lower_triangular:
            rows = slice(convex_comb, self._bottom_bound)
            cols = slice(self._left_bound, self._right_bound)
            return I[rows, :].tocsc()[:, cols].toarray()

        rows = slice(self._top_bound, convex_comb)
        cols = slice(self._left_bound, self._right_bound)
        return I[rows, :].tocsc()[:, cols].toarray()

    @staticmethod
    def _compute_inner_descriptors(I: NDArray) -> Tuple[NDArray[float], float, float]:
        return np.percentile(I, [0, 25, 50, 75, 100]), np.mean(I), np.std(I)

    def _compute_lmean(self, I: ss.csr_matrix, window: int) -> float:
        """
        Compute the mean intensity for the left neighborhood
        """
        assert window >= 0

        new_bound = max(0, self._left_bound - window)
        if new_bound == self._left_bound:
            return math.nan

        convex_comb = self._compute_convex_comp()
        if self.lower_triangular:
            submatrix = I[convex_comb : self._bottom_bound, new_bound : self._left_bound]
        else:
            submatrix = I[self._top_bound : convex_comb, new_bound : self._left_bound]

        return submatrix.mean()

    def _compute_rmean(self, I: ss.csr_matrix, window: int) -> float:
        """
        Compute the mean intensity for the right neighborhood
        """
        assert window >= 0

        new_bound = min(I.shape[1], self._right_bound + window)

        if new_bound == self._right_bound:
            return math.nan

        convex_comb = self._compute_convex_comp()
        if self.lower_triangular:
            submatrix = I[convex_comb : self._bottom_bound, self._right_bound : new_bound]
        else:
            submatrix = I[self._top_bound : convex_comb, self._right_bound : new_bound]

        return submatrix.mean()

    def _all_bounds_set(self) -> bool:
        return all((x is not None for x in [self._left_bound, self._right_bound, self._bottom_bound, self._top_bound]))

    @property
    def seed(self) -> int:
        return self._seed

    @property
    def top_persistence(self) -> Union[float, None]:
        return self._persistence

    @property
    def lower_triangular(self) -> bool:
        return self._where == "lower_triangular"

    @property
    def upper_triangular(self) -> bool:
        return self._where == "upper_triangular"

    @property
    def left_bound(self) -> int:
        if self._left_bound is None:
            raise RuntimeError("left_bound has not been set")
        return self._left_bound

    @property
    def right_bound(self) -> int:
        if self._right_bound is None:
            raise RuntimeError("right_bound has not been set")
        return self._right_bound

    @property
    def top_bound(self) -> int:
        if self._top_bound is None:
            raise RuntimeError("top_bound has not been set")
        return self._top_bound

    @property
    def bottom_bound(self) -> int:
        if self._bottom_bound is None:
            raise RuntimeError("bottom_bound has not been set")
        return self._bottom_bound

    @property
    def inner_mean(self) -> float:
        if self._inner_mean is None:
            raise RuntimeError(
                "caught an attempt to access inner_mean property before compute_biodescriptors() was called"
            )

        return self._inner_mean

    @property
    def inner_std(self) -> float:
        if self._inner_std is None:
            raise RuntimeError(
                "caught an attempt to access inner_std property before compute_biodescriptors() was called"
            )

        return self._inner_std

    @property
    def five_number(self) -> NDArray[float]:
        if self._five_number is None:
            raise RuntimeError(
                "caught an attempt to access five_number property before compute_biodescriptors() was called"
            )

        return self._five_number

    @property
    def outer_lmean(self) -> float:
        if self._outer_lmean is None:
            raise RuntimeError(
                "caught an attempt to access outer_lmean property before compute_biodescriptors() was called"
            )

        return self._outer_lmean

    @property
    def outer_rmean(self) -> float:
        if self._outer_rmean is None:
            raise RuntimeError(
                "caught an attempt to access outer_rmean property before compute_biodescriptors() was called"
            )

        return self._outer_rmean

    @property
    def outer_mean(self) -> float:
        if self.outer_rmean == -1:
            assert self.outer_lmean == -1
            return -1.0

        if math.isnan(self.outer_rmean):
            return self.outer_lmean

        if math.isnan(self.outer_lmean):
            return self.outer_rmean

        return (self.outer_lmean + self.outer_rmean) / 2

    @property
    def rel_change(self) -> float:
        if self.outer_mean <= 0:
            return -1.0

        return abs(self.inner_mean - self.outer_mean) / self.outer_mean * 100

    def set_horizontal_bounds(self, left_bound: int, right_bound: int):
        """
        Set the horizontal bounds for the stripe.
        This function raises an exception when the coordinates have already been set or when the
        given coordinates are incompatible with the seed position.

        Parameters
        ----------
        left_bound: int
        right_bound: int
        """
        if self._left_bound is not None:
            assert self._right_bound is not None
            raise RuntimeError("horizontal stripe bounds have already been set")

        if not left_bound <= self._seed <= right_bound:
            raise ValueError(
                f"horizontal bounds must enclose the seed position: seed={self._seed}, {left_bound=}, {right_bound=}"
            )

        self._left_bound = left_bound
        self._right_bound = right_bound

    def set_vertical_bounds(self, top_bound: int, bottom_bound: int):
        """
        Set the vertical bounds for the stripe.
        This function raises an exception when the coordinates have already been set or when the
        given coordinates are incompatible with the seed position and/or the where location.

        Parameters
        ----------
        top_bound: int
        bottom_bound: int
        """
        if self._bottom_bound is not None:
            assert self._top_bound is not None
            raise RuntimeError("vertical stripe bounds have already been set")

        if top_bound > bottom_bound:
            raise ValueError(
                f"the lower vertical bound must be greater than the upper vertical bound: {top_bound=}, {bottom_bound=}"
            )

        self._top_bound = top_bound
        self._bottom_bound = bottom_bound

        computed_where = self._infer_location(self._seed, self._top_bound, self._bottom_bound)

        if self._where is not None and computed_where != self._where:
            raise RuntimeError(
                f"computed location does not match the provided stripe location: computed={computed_where}, expected={self._where}"
            )

        self._where = computed_where

    def compute_biodescriptors(self, I: ss.csr_matrix, window: int = 3):
        """
        Use the sparse matrix I to compute various descriptive statistics.
        Statistics are stored in the current Stripe instance.
        This function raises an exception when it is called before the stripe bounds have been set.

        Parameters
        ----------
        I: ss.csr_matrix
            the sparse matrix from which the stripe originated
        window: int
            window size used to compute statistics to the left and right of the stripe
        """
        if not self._all_bounds_set():
            raise RuntimeError("compute_biodescriptors() was called on a bound-less stripe")

        if window < 0:
            raise ValueError("window cannot be negative")

        restrI = self._slice_matrix(I)

        # ATT This can avoid empty stripes, which can occur e.g. when the column has (approximately) constant entries
        if np.prod(restrI.size) == 0:
            self._five_number = np.array([-1.0] * 5)
            self._inner_mean = -1.0
            self._inner_std = -1.0
            self._outer_lmean = -1.0
            self._outer_rmean = -1.0
            return

        self._five_number, self._inner_mean, self._inner_std = self._compute_inner_descriptors(restrI)

        # Mean intensity - left and right neighborhoods:
        self._outer_lmean = self._compute_lmean(I, window)
        self._outer_rmean = self._compute_rmean(I, window)



# Copyright (C) 2024 Roberto Rossini <roberros@uio.no>
#
# SPDX-License-Identifier: MIT

import functools
import hashlib
import json
import logging
import math
import pathlib
import random
import sys
import tempfile
import time
import urllib.request
from typing import Any, Dict, Tuple, Union


@functools.cache
def _get_datasets(max_size: float) -> Dict[str, Dict[str, str]]:
    assert not math.isnan(max_size)

    record_id = "14517632"

    datasets = {
        "4DNFI3RFZLZ5": {
            "url": f"https://zenodo.org/records/{record_id}/files/4DNFI3RFZLZ5.stripepy.mcool?download=1",
            "md5": "f6e060211c95dd5fbf6e708c637d1c1c",
            "assembly": "mm10",
            "format": "mcool",
            "size_mb": 83.85,
        },
        "4DNFIC1CLPK7": {
            "url": f"https://zenodo.org/records/{record_id}/files/4DNFI6HDY7WZ.stripepy.mcool?download=1",
            "md5": "745df902a842c17e535222fb7f9748ca",
            "assembly": "hg38",
            "format": "mcool",
            "size_mb": 104.73,
        },
        "4DNFI9GMP2J8": {
            "url": f"https://zenodo.org/records/{record_id}/files/4DNFI9GMP2J8.stripepy.mcool?download=1",
            "md5": "a17d08460c03cf6c926e2ca5743e4888",
            "assembly": "hg38",
            "format": "mcool",
            "size_mb": 106.84,
        },
        "ENCFF993FGR": {
            "url": f"https://zenodo.org/records/{record_id}/files/ENCFF993FGR.stripepy.hic?download=1",
            "md5": "3bcb8c8c5aac237f26f994e0f5e983d7",
            "assembly": "hg38",
            "format": "hic",
            "size_mb": 185.29,
        },
        "__results_v1": {
            "url": f"https://zenodo.org/records/{record_id}/files/results_4DNFI9GMP2J8_v1.hdf5?download=1",
            "md5": "632b2a7a6e5c1a24dc3635710ed68a80",
            "filename": "results_4DNFI9GMP2J8_v1.hdf5",
            "assembly": "hg38",
            "format": "stripepy",
            "size_mb": 8.75,
        },
        "__stripepy_plot_images": {
            "url": f"https://zenodo.org/records/{record_id}/files/stripepy-plot-test-images.tar.xz?download=1",
            "md5": "d4ab74937dd9062efe4b2acc6ebc8780",
            "filename": "stripepy-plot-test-images.tar.xz",
            "assembly": "hg38",
            "format": "tar",
            "size_mb": 1.5,
        },
    }

    valid_dsets = {k: v for k, v in datasets.items() if v.get("size_mb", math.inf) < max_size}

    if len(valid_dsets) > 0:
        return valid_dsets

    raise RuntimeError(f"unable to find any dataset smaller than {max_size:.2f} MB")


def _list_datasets():
    dsets = {k: v for k, v in _get_datasets(math.inf).items() if not k.startswith("__")}
    json.dump(dsets, fp=sys.stdout, indent=2)
    sys.stdout.write("\n")


def _get_random_dataset(max_size: float) -> Tuple[str, Dict[str, str]]:
    dsets = _get_datasets(max_size)
    assert len(dsets) > 0

    key = random.sample(list(dsets.keys()), 1)[0]
    return key, dsets[key]


def _lookup_dataset(name: Union[str, None], assembly: Union[str, None], max_size: float) -> Tuple[str, Dict[str, str]]:
    if name is not None:
        max_size = math.inf
        try:
            return name, _get_datasets(max_size)[name]
        except KeyError as e:
            raise RuntimeError(
                f'unable to find dataset "{name}". Please make sure the provided dataset is present in the list produced by stripepy download --list-only.'
            ) from e

    assert assembly is not None
    assert max_size >= 0

    dsets = {k: v for k, v in _get_datasets(max_size).items() if v["assembly"] == assembly}
    if len(dsets) == 0:
        raise RuntimeError(
            f'unable to find a dataset using "{assembly}" as reference genome. Please make sure such dataset exists in the list produced by stripepy download --list-only.'
        )

    key = random.sample(list(dsets.keys()), 1)[0]
    return key, dsets[key]


def _hash_file(path: pathlib.Path, chunk_size=16 << 20) -> str:
    logging.info('computing MD5 digest for file "%s"...', path)
    with path.open("rb") as f:
        hasher = hashlib.md5()
        while True:
            chunk = f.read(chunk_size)
            if not chunk:
                return hasher.hexdigest()
            hasher.update(chunk)


def _download_progress_reporter(chunk_no, max_chunk_size, download_size):
    if download_size == -1:
        if not _download_progress_reporter.skip_progress_report:
            _download_progress_reporter.skip_progress_report = True
            logging.warning("unable to report download progress: remote file size is not known!")
        return

    timepoint = _download_progress_reporter.timepoint

    if time.time() - timepoint >= 15:
        mb_downloaded = (chunk_no * max_chunk_size) / (1024 << 10)
        download_size_mb = download_size / (1024 << 10)
        progress_pct = (mb_downloaded / download_size_mb) * 100
        logging.info("downloaded %.2f/%.2f MB (%.2f%%)", mb_downloaded, download_size_mb, progress_pct)
        _download_progress_reporter.timepoint = time.time()


# this is Python's way of defining static variables inside functions
_download_progress_reporter.skip_progress_report = False
_download_progress_reporter.timepoint = 0.0


def _download_and_checksum(name: str, dset: Dict[str, Any], dest: pathlib.Path):
    with tempfile.NamedTemporaryFile(dir=dest.parent, prefix=f"{dest.stem}.") as tmpfile:
        tmpfile.close()
        tmpfile = pathlib.Path(tmpfile.name)

        url = dset["url"]
        md5sum = dset["md5"]
        assembly = dset.get("assembly", "unknown")

        logging.info('downloading dataset "%s" (assembly=%s)...', name, assembly)
        t0 = time.time()
        urllib.request.urlretrieve(url, tmpfile, reporthook=_download_progress_reporter)
        t1 = time.time()
        logging.info('DONE! Downloading dataset "%s" took %.2fs.', name, t1 - t0)

        digest = _hash_file(tmpfile)
        if digest == md5sum:
            logging.info("MD5 checksum match!")
            return tmpfile.rename(dest)

        raise RuntimeError(
            f'MD5 checksum for file downloaded from "{url}" does not match: expected {md5sum}, found {digest}.'
        )


def run(
    name: Union[str, None],
    output_path: Union[pathlib.Path, None],
    assembly: Union[str, None],
    max_size: float,
    list_only: bool,
    force: bool,
):
    t0 = time.time()
    if list_only:
        _list_datasets()
        return

    do_random_sample = name is None and assembly is None

    if do_random_sample:
        dset_name, config = _get_random_dataset(max_size)
    else:
        dset_name, config = _lookup_dataset(name, assembly, max_size)

    if output_path is None:
        if "filename" in config:
            output_path = pathlib.Path(config["filename"])
        else:
            output_path = pathlib.Path(f"{dset_name}.{config['format']}")

    if output_path.exists() and not force:
        raise RuntimeError(f"refusing to overwrite file {output_path}. Pass --force to overwrite.")
    output_path.unlink(missing_ok=True)

    dest = _download_and_checksum(dset_name, config, output_path)
    t1 = time.time()

    logging.info('successfully downloaded dataset "%s" to file "%s"', config["url"], dest)
    logging.info(f"file size: %.2fMB. Elapsed time: %.2fs", dest.stat().st_size / (1024 << 10), t1 - t0)



# Copyright (C) 2024 Roberto Rossini <roberros@uio.no>
#
# SPDX-License-Identifier: MIT

import functools
import warnings
from typing import Any, Dict, List, Optional, Sequence, Tuple

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
from matplotlib.image import AxesImage
from matplotlib.ticker import EngFormatter, ScalarFormatter


@functools.cache
def _get_custom_palettes() -> Dict[str, npt.NDArray]:
    # Source: https://github.com/open2c/cooltools/blob/master/cooltools/lib/plotting.py
    return {
        "fall": np.array(
            (
                (255, 255, 255),
                (255, 255, 204),
                (255, 237, 160),
                (254, 217, 118),
                (254, 178, 76),
                (253, 141, 60),
                (252, 78, 42),
                (227, 26, 28),
                (189, 0, 38),
                (128, 0, 38),
                (0, 0, 0),
            )
        )
        / 255,
        "fruit_punch": np.array(
            (
                (255, 255, 255),
                (255, 204, 204),
                (255, 153, 153),
                (255, 102, 102),
                (255, 50, 50),
                (255, 0, 0),
            )
        )
        / 255,
        "blues": np.array(
            (
                (255, 255, 255),
                (180, 204, 225),
                (116, 169, 207),
                (54, 144, 192),
                (5, 112, 176),
                (4, 87, 135),
                (3, 65, 100),
                (2, 40, 66),
                (1, 20, 30),
                (0, 0, 0),
            )
        )
        / 255,
        "acidblues": np.array(
            (
                (255, 255, 255),
                (162, 192, 222),
                (140, 137, 187),
                (140, 87, 167),
                (140, 45, 143),
                (120, 20, 120),
                (90, 15, 90),
                (60, 10, 60),
                (30, 5, 30),
                (0, 0, 0),
            )
        )
        / 255,
        "nmeth": np.array(
            (
                (236, 250, 255),
                (148, 189, 217),
                (118, 169, 68),
                (131, 111, 43),
                (122, 47, 25),
                (41, 0, 20),
            )
        )
        / 255,
    }


def _list_to_colormap(color_list, name=None) -> mpl.colors.LinearSegmentedColormap:
    color_list = np.array(color_list)
    if color_list.min() < 0:
        raise ValueError("Colors should be 0 to 1, or 0 to 255")
    if color_list.max() > 1.0:
        if color_list.max() > 255:
            raise ValueError("Colors should be 0 to 1 or 0 to 255")
        else:
            color_list = color_list / 255.0
    return mpl.colors.LinearSegmentedColormap.from_list(name, color_list, 256)


def _register_cmaps():
    # make sure we are not trying to register color maps multiple times
    if hasattr(_register_cmaps, "called"):
        return
    for name, pal in _get_custom_palettes().items():
        mpl.colormaps.register(_list_to_colormap(pal, name))
        mpl.colormaps.register(_list_to_colormap(pal[::-1], f"{name}_r"))

    _register_cmaps.called = True


def _format_ticks(ax: plt.Axes, xaxis: bool = True, yaxis: bool = True, rotation: int = 30):
    """
    Function taken from https://cooltools.readthedocs.io/en/latest/notebooks/viz.html
    """
    if xaxis:
        ax.xaxis.set_major_formatter(EngFormatter("b"))
    else:
        ax.xaxis.set_major_formatter(ScalarFormatter())
    ax.xaxis.tick_bottom()

    if yaxis:
        ax.yaxis.set_major_formatter(EngFormatter("b"))
    else:
        ax.yaxis.set_major_formatter(ScalarFormatter())

    if rotation != 0:
        ax.tick_params(axis="x", rotation=rotation)


def hic_matrix(
    matrix: npt.NDArray,
    region: Tuple[int, int],
    cmap="fruit_punch",
    log_scale: bool = True,
    with_colorbar: bool = False,
    fig: Optional[plt.Figure] = None,
    ax: Optional[plt.Axes] = None,
) -> Tuple[plt.Figure, plt.Axes, AxesImage]:
    """
    Plot the given matrix as a heatmap

    Parameters
    ----------
    matrix: npt.NDArray
        the 2D matrix to be plotted
    region: Tuple[int, int]
        a 2 int tuple containing the first and last genomic coordinates (bp) to be plotted
    cmap
        color map to be used for plotting
    log_scale: bool
        if True, plot the heatmap in log scale
    with_colorbar: bool
        if True, add a color bar for the plotted heatmap
    fig: Optional[plt.Figure]
        figure to use for plotting
    ax: Optional[plt.Axes]
        axis to use for plotting

    Returns
    -------
    fig: plt.Figure
        figure used for plotting
    ax: plt.Axes
        axis used for plotting
    img:
        image returned by ax.imshow()
    """
    _register_cmaps()

    if fig is None:
        if ax is not None:
            raise RuntimeError("ax should be None when fig is None")
        fig, ax = plt.subplots(1, 1)
    elif ax is None:
        raise RuntimeError("ax cannot be None when fig is not None")

    kwargs = {
        "extent": (region[0], region[1], region[1], region[0]),
        "cmap": cmap,
    }
    if log_scale:
        kwargs["norm"] = mpl.colors.LogNorm()
    else:
        kwargs["vmax"] = np.amax(matrix)

    img = ax.imshow(matrix, **kwargs)
    _format_ticks(ax)

    if with_colorbar:
        fig.colorbar(img, ax=ax)

    return fig, ax, img


def pseudodistribution(
    pseudo_distrib_lt: Sequence[float],
    pseudo_distrib_ut: Sequence[float],
    region: Tuple[int, int],
    resolution: int,
    highlighted_points_lt: Optional[Sequence[int]] = None,
    highlighted_points_ut: Optional[Sequence[int]] = None,
    colors: Optional[Any] = None,
    fig: Optional[plt.Figure] = None,
    axs: Optional[Tuple[plt.Axes, plt.Axes]] = None,
) -> Tuple[plt.Figure, Tuple[plt.Axes, plt.Axes]]:
    """
    Plot the given pseudo-distributions as a two line plots

    Parameters
    ----------
    pseudo_distrib_lt: Sequence[float]
        pseudo-distribution values computed on the lower-triangular matrix
    pseudo_distrib_ut: Sequence[float]
        pseudo-distribution values computed on the upper-triangular matrix
    region: Tuple[int, int]
        a 2 int tuple containing the first and last genomic coordinates (bp) to be plotted
    resolution: int
        resolution (bp) of the Hi-C matrix from which the pseudo-distribution were computed
    highlighted_points_lt: Optional[Sequence[int]]
        list of genomic coordinates (lower-triangular matrix) to be highlighted.
        Each pair of genomic coordinates determines a point cloud as follows:
        - for genomic coordinate x, we sample the value of pseudo_distrib_lt at x
        - each point cloud is scatterplotted with the color specified through the colors parameter.
    highlighted_points_ut: Optional[Sequence[int]]
        same as coords2scatter_lt but for the upper-triangular matrix
    colors: Optional[Any]
        one or more colors to be used when generating the scatter plot.
        When provided it should be either:
        - a single color: all points are plotted using the same color
        - a list colors: the list should contain at least
          max(len(color2scatter_lt), len(color2scatter_ut)) colors
    fig: Optional[plt.Figure]
        figure to use for plotting
    axs: Optional[Tuple[plt.Axes, plt.Axes]]
        axes to use for plotting

    Returns
    -------
    fig: plt.Figure
        figure used for plotting
    ax1: plt.Axes
        axis with the plot for upper-triangular pseudo-distribution
    ax2: plt.Axes
        axis with the plot for lower-triangular pseudo-distribution
    """
    if fig is None:
        if axs is not None:
            raise RuntimeError("axs should be None when fig is None")
        fig, axs = plt.subplots(2, 1, sharex=True)
    elif axs is None:
        raise RuntimeError("axs cannot be None when fig is not None")

    if len(axs) != 2:
        raise RuntimeError("axs should be a tuple with exactly two plt.Axes")

    i1 = region[0] // resolution
    i2 = region[1] // resolution

    if colors is None:
        num_colors = 0
        if highlighted_points_lt is not None:
            num_colors = len(highlighted_points_lt)
        if highlighted_points_ut is not None:
            num_colors = max(num_colors, len(highlighted_points_ut))

        if num_colors > 0:
            colors = ["blue"] * num_colors

    def _plot(data, points, ax, title):
        ax.plot(
            np.arange(i1, i2) * resolution,
            data[i1:i2],
            color="red",
            linewidth=0.5,
            linestyle="solid",
        )
        if points is not None:
            for n, x in enumerate(points):
                if not region[0] <= x <= region[1]:
                    raise RuntimeError("point outside of bounds")
                ax.plot(
                    x,
                    data[x // resolution],
                    marker=".",
                    linestyle="",
                    markersize=5,
                    color=colors[n],
                )
        ax.set(title=title, ylim=(0, 1), ylabel="Pseudo-distribution")
        ax.grid(True)

    _plot(pseudo_distrib_ut, highlighted_points_ut, axs[0], "Upper Triangular")
    _plot(pseudo_distrib_lt, highlighted_points_lt, axs[1], "Lower Triangular")

    _format_ticks(axs[1], yaxis=False, xaxis=True)

    axs[1].set(xlabel="Genomic coordinates (bp)")

    return fig, axs


def plot_sites(
    sites: Sequence[int],
    region: Tuple[int, int],
    location: Optional[str],
    color="blue",
    fig: Optional[plt.Figure] = None,
    ax: Optional[plt.Axes] = None,
):
    """
    Plot one or more sites of interest (e.g. stipe seeds).

    Parameters
    ----------
    sites: Sequence[int]
        list of genomic coordinates (bp) corresponding to the x-coordinates of the sites to be plotted
    region: Tuple[int, int]
        a 2 int tuple containing the first and last genomic coordinates (bp) to be plotted
    location: Optional[str]
        if "lower" (resp. "upper"), then plot sites only on the lower (resp. upper) part of
        the Hi-C matrix. Otherwise, plot sites spanning the whole Hi-C matrix
    color:
        color to be used for plotting
    fig: Optional[plt.Figure]
        figure to use for plotting
    ax: Optional[Tuple[plt.Axes, plt.Axes]]
        axis to use for plotting

    Returns
    -------
    fig: plt.Figure
        figure used for plotting
    ax: plt.Axes
        axis used for plotting
    """
    if location is not None and location not in {"lower", "upper"}:
        raise ValueError('location should be "lower" or "upper"')

    if fig is None:
        if ax is not None:
            raise RuntimeError("ax should be None when fig is None")
        fig, ax = plt.subplots(1, 1)
    elif ax is None:
        raise RuntimeError("ax cannot be None when fig is not None")

    if len(sites) == 0:
        warnings.warn("stripepy.plot.plot_sites: no sites to plot!")

    for site in sites:
        if not region[0] <= site <= region[1]:
            raise RuntimeError("site outside of bounds")

        x = [site] * 2
        if location == "lower":
            y = [site, region[1] - 1]
        elif location == "upper":
            y = [region[0] + 1, site]
        else:
            y = [region[0] + 1, region[1] - 1]

        ax.plot(x, y, color=color, linestyle="dashed", linewidth=1)

    _format_ticks(ax)

    return fig, ax


def mask_regions_1d(
    matrix: npt.NDArray,
    resolution: int,
    location: Optional[str],
    whitelist: Optional[List[Tuple[int, int]]] = None,
    blacklist: Optional[List[Tuple[int, int]]] = None,
) -> npt.NDArray:
    """
    Mask rows or columns of the given matrix based on whitelisted or blacklisted regions.

    Parameters
    ----------
    matrix: npt.NDArray
        matrix to be masked. The matrix is expected to be a 2D matrix and be symmetric.
    resolution: int
        matrix resolution in bp.
    location: str
        location where the selective masking should be applied.
        When "lower", values in the upper triangular matrix are all set to 0 and values
        in the lower triangular matrix are masked based on whitelisted or blacklisted regions.
        When "upper", values in the lower triangular matrix are all set to 0 and values
        in the upper triangular matrix are masked based on whitelisted or blacklisted regions.
    whitelist: Optional[List[Tuple[int, int]]]
        list of regions to NOT be masked. Regions should be expressed in matrix coordinates.
    blacklist: Optional[List[Tuple[int, int]]]
        list of regions to be masked. Regions should be expressed in matrix coordinates.

    Returns
    -------
    matrix: npt.NDArray
        the masked matrix
    """
    if location is not None and location not in {"lower", "upper"}:
        raise ValueError('location should be "lower" or "upper"')

    if (whitelist is None and blacklist is None) or (whitelist is not None and blacklist is not None):
        raise ValueError("please specify either whitelist or blacklist")

    m = matrix.copy()
    mask = None
    if whitelist is not None:
        idx = []
        if len(whitelist) > 0:
            for i1, i2 in whitelist:
                i1 //= resolution
                i2 //= resolution
                idx.extend(list(range(i1, i2 + 1)))

        mask = np.setdiff1d(np.arange(m.shape[0]), np.unique(idx))

    if blacklist is not None:
        mask = []
        if len(blacklist) > 0:
            for i1, i2 in blacklist:
                i1 //= resolution
                i2 //= resolution
                mask.extend(list(range(i1, i2 + 1)))

        mask = np.unique(mask)

    if mask is not None:
        m[:, mask] = 0

    if location == "upper":
        return np.triu(m)
    if location == "lower":
        return np.tril(m)
    return np.tril(m.T) + np.triu(m)


def draw_boxes(
    regions: List[Tuple[int, int, int, int]],
    bound_box: Tuple[int, int],
    fig: Optional[plt.Figure] = None,
    ax: Optional[plt.Axes] = None,
    **kwargs,
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Draw bounding boxes around a list of regions of interest.

    Parameters
    ----------
    regions:
        a list of genomic regions, each consisting of a 4-element tuple (x, y, width, height).
        All coordinates are in genomic units (i.e. bp).
    bound_box:
        first and last genomic position that should be plotted.
        Used to ensure that boxes do not extend outside the plotted area.
    fig:
        figure to use for plotting
    ax:
        axis to use for plotting
    **kwargs:
        additional keyword arguments to pass to plt.plot()

    Returns
    -------
    figure, axes:
        the plt.Figure and plt.Axes used for plotting
    """
    if fig is None:
        if ax is not None:
            raise RuntimeError("ax should be None when fig is None")
        fig, ax = plt.subplots(1, 1)
    elif ax is None:
        raise RuntimeError("ax cannot be None when fig is not None")

    if "color" not in kwargs:
        kwargs["color"] = "blue"
    if "linestyle" not in kwargs:
        kwargs["linestyle"] = "dashed"

    for x, y, width, height in regions:
        x = np.clip([x, x, x + width, x + width, x], bound_box[0], bound_box[1])
        y = np.clip([y, y + height, y + height, y + width, y], bound_box[0], bound_box[1])
        ax.plot(x, y, **kwargs)
    return fig, ax



# Copyright (C) 2024 Roberto Rossini <roberroso@uio.no>
#
# SPDX-License-Identifier: MIT

import contextlib
import io
import pathlib

import pandas as pd
import pytest

from stripepy import main

testdir = pathlib.Path(__file__).resolve().parent.parent


@pytest.mark.end2end
class TestStripePyView:
    @staticmethod
    def setup_class():
        test_files = [testdir / "data" / "results_4DNFI9GMP2J8_v1.hdf5"]

        for f in test_files:
            if not f.exists():
                raise RuntimeError(
                    f'unable to find file "{f}". Did you download the test files prior to running pytest?'
                )

    @staticmethod
    def test_view():
        testfile = testdir / "data" / "results_4DNFI9GMP2J8_v1.hdf5"

        args = ["view", str(testfile)]
        buff = io.StringIO()
        with contextlib.redirect_stdout(buff):
            main(args)

        buff.seek(0)
        cols = ["chrom1", "start1", "end1", "chrom2", "start2", "end2"]
        df = pd.read_table(buff, names=cols)

        assert len(df.columns) == len(cols)
        assert len(df) == len(df[~df.isnull().values])
        assert len(df) == 19750



# Copyright (C) 2024 Roberto Rossini <roberros@uio.no>
#
# SPDX-License-Identifier: MIT

import itertools
import logging
import pathlib
import random
import warnings
from typing import Dict, Optional, Tuple

import hictkpy
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.ticker import EngFormatter
from numpy.typing import NDArray

import stripepy.plot
from stripepy.IO import ResultFile
from stripepy.utils.TDA import TDA


def _generate_random_region(
    chroms: Dict[str, int], resolution: int, region_size: int = 2_500_000
) -> Tuple[str, int, int]:
    chrom_names = list(chroms.keys())
    random.shuffle(chrom_names)

    for chrom in chrom_names:
        size = chroms[chrom]
        if size < region_size:
            continue

        if size == region_size:
            return (
                chrom,
                0,
                size,
            )

        offset = region_size // 2
        pos = random.randint(offset, chroms[chrom] - offset)
        start_pos = ((pos - offset) // resolution) * resolution
        end_pos = start_pos + region_size
        return (
            chrom,
            start_pos,
            end_pos,
        )

    raise RuntimeError(
        "Unable to randomly select a region to be plotted. Please manually select the desired region by passing the --region option."
    )


def _fetch_random_region(
    f: hictkpy.File,
    normalization: Optional[str],
    region_size: int = 2_500_000,
) -> Tuple[str, int, int, NDArray]:
    for attempt in range(10):
        chrom, start_pos, end_pos = _generate_random_region(
            f.chromosomes(include_ALL=False), f.resolution(), region_size
        )
        m = f.fetch(f"{chrom}:{start_pos}-{end_pos}", normalization=normalization).to_numpy()
        nnz = (np.isfinite(m) & (m != 0)).sum()

        if nnz / m.size >= 0.75:
            return chrom, start_pos, end_pos, m

    warnings.warn(
        "Failed to randomly select a genomic region with appropriate density for plotting.\n"
        "Continuing anyway.\n"
        "For best results, please manually provide the coordinates for a region to be plotted using parameter --region."
    )

    return chrom, start_pos, end_pos, m  # noqa


def _fetch_matrix(
    path: pathlib.Path, resolution: int, normalization: str, region: Optional[str]
) -> Tuple[str, int, int, NDArray]:
    f = hictkpy.MultiResFile(path)[resolution]

    if region is None:
        return _fetch_random_region(f, normalization)

    matrix = f.fetch(region, normalization=normalization).to_numpy()
    chrom, start, end = _parse_ucsc_region(region, f.chromosomes(include_ALL=False))

    return chrom, start, end, matrix


def _parse_ucsc_region(region: str, chromosomes: Dict[str, int]) -> Tuple[str, int, int]:
    try:
        chrom, sep, pos = region.partition(":")
        if len(sep) == 0 and len(pos) == 0:
            return chrom, 0, chromosomes[chrom]

        start_pos, _, end_pos = pos.partition("-")
        return chrom, int(start_pos.replace(",", "")), int(end_pos.replace(",", ""))
    except Exception as e:
        raise RuntimeError(f'Unable to parse region "{region}". Is the given region in UCSC format?') from e


def _validate_hdf5_result(hf: hictkpy.File, rf: ResultFile):
    if hf.resolution() != rf.resolution:
        raise RuntimeError(f'File "{hf.uri()}" and "{rf.path}" have different resolutions')

    if hf.chromosomes(include_ALL=False) != rf.chromosomes:
        raise RuntimeError(f'File "{hf.uri()}" and "{rf.path}" have different chromosomes')


def _fetch_geo_descriptors(
    h5: ResultFile,
    chrom: str,
    left_bound: int,
    right_bound: int,
    location: str,
) -> pd.DataFrame:
    assert location in {"LT", "UT"}
    assert left_bound >= 0
    assert right_bound >= left_bound

    df = h5.get(chrom, "geo_descriptors", location)

    for col in df.columns:
        if col == "top_persistence":
            continue
        df[col] = np.minimum(df[col] * h5.resolution, h5.chromosomes[chrom])

    return df[df["seed"].between(left_bound, right_bound, inclusive="both")]


def _fetch_geo_descriptors_gw(
    h5: ResultFile,
    location: str,
) -> pd.DataFrame:
    assert location in {"LT", "UT"}

    dfs = []
    for chrom in h5.chromosomes:
        try:
            df = h5.get(chrom, "geo_descriptors", location)
            for col in df.columns:
                if col == "top_persistence":
                    continue
                df[col] = np.minimum(df[col] * h5.resolution, h5.chromosomes[chrom])
            dfs.append(df)

        except KeyError:
            pass

    return pd.concat(dfs).reset_index(drop=True)


def _fetch_persistence_maximum_points(h5: ResultFile, chrom: str, start: int, end: int) -> Dict[str, NDArray]:
    def fetch(v: NDArray[int], left_bound: int, right_bound: int) -> Tuple[NDArray[int], NDArray[int]]:
        assert left_bound >= 0
        assert right_bound >= left_bound

        idx = np.where((v >= left_bound) & (v < right_bound))[0]
        return idx.astype(int), v[idx].astype(int)

    pd_lt = h5.get(chrom, "pseudodistribution", "LT")["pseudodistribution"].to_numpy()
    pd_ut = h5.get(chrom, "pseudodistribution", "UT")["pseudodistribution"].to_numpy()

    min_persistence = h5.get_min_persistence(chrom)
    lt_idx, lt_seeds = fetch(
        np.sort(TDA(pd_lt, min_persistence=min_persistence)[2]), start // h5.resolution, end // h5.resolution
    )
    ut_idx, ut_seeds = fetch(
        np.sort(TDA(pd_ut, min_persistence=min_persistence)[2]), start // h5.resolution, end // h5.resolution
    )

    return {
        "pseudodistribution_lt": pd_lt,
        "pseudodistribution_ut": pd_ut,
        "seeds_lt": lt_seeds,
        "seeds_ut": ut_seeds,
        "seed_indices_lt": lt_idx,
        "seed_indices_ut": ut_idx,
    }


def _plot_hic_matrix(
    contact_map: pathlib.Path,
    resolution: int,
    region: Optional[str],
    cmap: str,
    normalization: str,
    log_scale: bool,
    **kwargs,
) -> plt.Figure:
    chrom, start, end, matrix = _fetch_matrix(contact_map, resolution, normalization, region)

    fig, _, _ = stripepy.plot.hic_matrix(
        matrix,
        (start, end),
        cmap=cmap,
        log_scale=log_scale,
        with_colorbar=True,
    )

    fig.suptitle(f"{chrom}:{start}-{end}")
    fig.tight_layout()

    return fig


def _plot_hic_matrix_with_seeds(
    contact_map: pathlib.Path,
    stripepy_hdf5: pathlib.Path,
    resolution: int,
    region: Optional[str],
    cmap: str,
    normalization: str,
    log_scale: bool,
    **kwargs,
) -> plt.Figure:
    chrom, start, end, matrix = _fetch_matrix(contact_map, resolution, normalization, region)

    with ResultFile(stripepy_hdf5) as h5:
        data = _fetch_persistence_maximum_points(h5, chrom, start, end)
        resolution = h5.resolution

    fig, axs = plt.subplots(1, 2, figsize=(12.8, 6.6), sharey=True)

    for ax in axs:
        _, _, img = stripepy.plot.hic_matrix(
            matrix,
            (start, end),
            cmap=cmap,
            log_scale=log_scale,
            with_colorbar=False,
            fig=fig,
            ax=ax,
        )

    stripepy.plot.plot_sites(
        data["seeds_lt"] * resolution,
        (start, end),
        location="lower",
        fig=fig,
        ax=axs[0],
    )

    stripepy.plot.plot_sites(
        data["seeds_ut"] * resolution,
        (start, end),
        location="upper",
        fig=fig,
        ax=axs[1],
    )

    fig.suptitle(f"{chrom}:{start}-{end}")
    fig.tight_layout()

    fig.subplots_adjust(right=0.94)
    cbar_ax = fig.add_axes((0.95, 0.15, 0.015, 0.7))
    fig.colorbar(img, cax=cbar_ax)  # noqa

    return fig


def _plot_hic_matrix_with_stripes(
    contact_map: pathlib.Path,
    stripepy_hdf5: pathlib.Path,
    resolution: int,
    relative_change_threshold: Optional[float],
    region: Optional[str],
    cmap: str,
    normalization: str,
    override_height: Optional[int],
    mask_regions: bool,
    log_scale: bool,
    **kwargs,
) -> plt.Figure:
    chrom, start, end, matrix = _fetch_matrix(contact_map, resolution, normalization, region)

    with ResultFile(stripepy_hdf5) as h5:
        chrom_size = h5.chromosomes[chrom]
        geo_descriptors_lt = _fetch_geo_descriptors(h5, chrom, start, end, "LT")
        geo_descriptors_ut = _fetch_geo_descriptors(h5, chrom, start, end, "UT")
        if relative_change_threshold is not None:
            mask_lt = (
                h5.get(chrom, "bio_descriptors", "LT")["rel_change"].iloc[geo_descriptors_lt.index]
                >= relative_change_threshold
            )
            mask_ut = (
                h5.get(chrom, "bio_descriptors", "UT")["rel_change"].iloc[geo_descriptors_ut.index]
                >= relative_change_threshold
            )

            geo_descriptors_lt = geo_descriptors_lt[mask_lt]
            geo_descriptors_ut = geo_descriptors_ut[mask_ut]

    fig, axs = plt.subplots(1, 2, figsize=(12.8, 6.6), sharey=True)

    outlines_lt = [
        (min(lb - start, chrom_size), min(rb - start, chrom_size), min(bb - tb, chrom_size))
        for lb, rb, bb, tb in geo_descriptors_lt[["left_bound", "right_bound", "bottom_bound", "top_bound"]].itertuples(
            index=False
        )
    ]

    outlines_ut = [
        (min(lb - start, chrom_size), min(rb - start, chrom_size), min(tb - bb, chrom_size))
        for lb, rb, bb, tb in geo_descriptors_ut[["left_bound", "right_bound", "bottom_bound", "top_bound"]].itertuples(
            index=False
        )
    ]

    if mask_regions:
        whitelist = [(x, y) for x, y, _ in outlines_lt]
        m1 = np.triu(matrix) + np.tril(
            stripepy.plot.mask_regions_1d(
                matrix,
                resolution,
                whitelist=whitelist,
                location="lower",
            ),
            k=1,
        )

        whitelist = [(x, y) for x, y, _ in outlines_ut]
        m2 = np.tril(matrix) + np.triu(
            stripepy.plot.mask_regions_1d(
                matrix,
                resolution,
                whitelist=whitelist,
                location="upper",
            ),
            k=1,
        )
    else:
        m1 = matrix
        m2 = matrix

    _, _, img = stripepy.plot.hic_matrix(
        m1,
        (start, end),
        cmap=cmap,
        log_scale=log_scale,
        with_colorbar=False,
        fig=fig,
        ax=axs[0],
    )
    stripepy.plot.hic_matrix(
        m2,
        (start, end),
        cmap=cmap,
        log_scale=log_scale,
        with_colorbar=False,
        fig=fig,
        ax=axs[1],
    )

    rectangles = []
    for lb, ub, height in outlines_lt:
        x = min(start + lb, chrom_size)
        y = min(start + lb, chrom_size)
        width = min(ub - lb, h5.chromosomes[chrom])
        if override_height is not None:
            height = min(end - x, override_height)
        rectangles.append((x, y, width, height))

    stripepy.plot.draw_boxes(rectangles, (start, end), color="blue", linestyle="dashed", fig=fig, ax=axs[0])

    rectangles = []
    for lb, ub, height in outlines_ut:
        x = min(start + lb, chrom_size)
        y = min(start + lb, chrom_size)
        width = min(ub - lb, chrom_size)
        if override_height is not None:
            height = min(start - x, override_height)
        rectangles.append((x, y, width, height))

    stripepy.plot.draw_boxes(rectangles, (start, end), color="blue", linestyle="dashed", fig=fig, ax=axs[1])

    axs[0].set(title="Lower Triangular")
    axs[1].set(title="Upper Triangular")

    fig.suptitle(f"{chrom}:{start}-{end}")
    fig.tight_layout()

    fig.subplots_adjust(right=0.94)
    cbar_ax = fig.add_axes((0.95, 0.15, 0.015, 0.7))
    fig.colorbar(img, cax=cbar_ax)

    return fig


def _plot_pseudodistribution(stripepy_hdf5: pathlib.Path, region: Optional[str], **kwargs) -> plt.Figure:
    with ResultFile(stripepy_hdf5) as h5:
        if region is None:
            chrom, start, end = _generate_random_region(h5.chromosomes, h5.resolution)
        else:
            chrom, start, end = _parse_ucsc_region(region, h5.chromosomes)
        data = _fetch_persistence_maximum_points(h5, chrom, start, end)
        resolution = h5.resolution

    fig, _ = stripepy.plot.pseudodistribution(
        data["pseudodistribution_lt"],
        data["pseudodistribution_ut"],
        region=(start, end),
        resolution=resolution,
        highlighted_points_lt=data["seeds_lt"] * resolution,
        highlighted_points_ut=data["seeds_ut"] * resolution,
    )

    fig.tight_layout()
    return fig


def _plot_stripe_dimension_distribution(
    stripepy_hdf5: pathlib.Path,
    region: Optional[str],
    **kwargs,
) -> plt.Figure:
    with ResultFile(stripepy_hdf5) as h5:
        if region is None:
            geo_descriptors_lt = _fetch_geo_descriptors_gw(h5, "LT")
            geo_descriptors_ut = _fetch_geo_descriptors_gw(h5, "UT")
        else:
            chrom, start, end = _parse_ucsc_region(region, h5.chromosomes)
            geo_descriptors_lt = _fetch_geo_descriptors(h5, chrom, start, end, "LT")
            geo_descriptors_ut = _fetch_geo_descriptors(h5, chrom, start, end, "UT")

        resolution = h5.resolution

    fig, axs = plt.subplots(2, 2, figsize=(12.8, 8), sharex="col", sharey="col")

    stripe_widths_lt = geo_descriptors_lt["right_bound"] - geo_descriptors_lt["left_bound"]
    stripe_heights_lt = geo_descriptors_lt["bottom_bound"] - geo_descriptors_lt["top_bound"]

    stripe_widths_ut = geo_descriptors_ut["right_bound"] - geo_descriptors_ut["left_bound"]
    stripe_heights_ut = geo_descriptors_ut["bottom_bound"] - geo_descriptors_ut["top_bound"]

    for ax in itertools.chain.from_iterable(axs):
        ax.xaxis.set_major_formatter(EngFormatter("b"))
        ax.xaxis.tick_bottom()

    axs[0][0].hist(stripe_widths_lt, bins=max(1, (stripe_widths_lt.max() - stripe_widths_lt.min()) // resolution))
    axs[0][1].hist(stripe_heights_lt, bins="auto")
    axs[1][0].hist(stripe_widths_ut, bins=max(1, (stripe_widths_ut.max() - stripe_widths_ut.min()) // resolution))
    axs[1][1].hist(stripe_heights_ut, bins="auto")

    axs[0][0].set(title="Stripe width distribution (lower triangle)", ylabel="Count")
    axs[0][1].set(title="Stripe height distribution (lower triangle)")
    axs[1][0].set(title="Stripe width distribution (upper triangle)", xlabel="Width (bp)", ylabel="Count")
    axs[1][1].set(title="Stripe height distribution (upper triangle)", xlabel="Height (bp)")

    if region is not None:
        fig.suptitle(f"{chrom}:{start}-{end}")

    fig.tight_layout()
    return fig


def run(plot_type: str, output_name: pathlib.Path, dpi: int, force: bool, **kwargs):
    if output_name.exists():
        if force:
            logging.debug('removing existing output file "%s"', output_name)
            output_name.unlink()
        else:
            raise RuntimeError(
                f'Refusing to overwrite file "{output_name}". Pass --force to overwrite existing file(s).'
            )

    if "seed" in kwargs:
        random.seed(kwargs["seed"])

    if plot_type in {"contact-map", "cm"}:
        plot_seeds = kwargs.pop("highlight_seeds")
        plot_stripes = kwargs.pop("highlight_stripes")
        ignore_stripe_heights = kwargs.pop("ignore_stripe_heights")

        if (plot_seeds or plot_stripes) and kwargs["stripepy_hdf5"] is None:
            raise RuntimeError("--stripepy-hdf5 is required when highlighting stripes or seeds.")

        if not plot_seeds and not plot_stripes:
            fig = _plot_hic_matrix(**kwargs)
        elif plot_seeds:
            fig = _plot_hic_matrix_with_seeds(**kwargs)
        else:
            if ignore_stripe_heights:
                kwargs["override_height"] = 2_500_000
                kwargs["mask_regions"] = True
            else:
                kwargs["override_height"] = None
                kwargs["mask_regions"] = False
            fig = _plot_hic_matrix_with_stripes(**kwargs)
    elif plot_type in {"pseudodistribution", "pd"}:
        fig = _plot_pseudodistribution(**kwargs)
    elif plot_type in {"stripe-hist", "hist"}:
        fig = _plot_stripe_dimension_distribution(**kwargs)
    else:
        raise NotImplementedError

    fig.savefig(output_name, dpi=dpi)



# Copyright (C) 2024 Andrea Raffo <andrea.raffo@ibv.uio.no>
#
# SPDX-License-Identifier: MIT

import numpy as np
import pytest
import scipy.sparse as ss

from stripepy.stripepy import _log_transform


@pytest.mark.unit
class TestLogTransform:

    def test_empty(self):
        I = ss.csr_matrix((0, 0), dtype=float)
        Iproc = _log_transform(I)

        assert Iproc.shape == (0, 0)
        assert np.issubdtype(I.dtype, np.floating)

    def test_all_finite(self):
        I = ss.rand(100, 100, density=0.5, format="csr")
        Iproc = _log_transform(I)

        assert I.size == Iproc.size
        assert I.shape == Iproc.shape
        assert np.isfinite(Iproc.data).all()

    def test_with_nans(self):
        I = ss.rand(100, 100, density=0.5, format="csr")
        size_I = I.size
        mean_I = I.mean()
        num_nan_values = (I.data >= mean_I).sum()
        I.data[I.data >= mean_I] = np.nan

        Iproc = _log_transform(I)

        assert np.isfinite(Iproc.data).all()
        assert Iproc.size == size_I - num_nan_values



# Copyright (C) 2024 Andrea Raffo <andrea.raffo@ibv.uio.no>
#
# SPDX-License-Identifier: MIT

import numpy as np

from .persistence1d import (
    diversify_extremum_points_and_persistence,
    filter_extremum_points_by_persistence,
    run_persistence,
)


def TDA(marginal_pd, min_persistence=0):

    # Compute the extremum points (i.e., minimum and maximum points) and their persistence:
    extremum_points_and_persistence = run_persistence(marginal_pd, level_sets="upper")

    # Filter extremum points by persistence threshold:
    filtered_extremum_points_and_persistence = filter_extremum_points_by_persistence(
        extremum_points_and_persistence, threshold=min_persistence
    )

    # Split extremum points into minimum and maximum points:
    (filtered_min_points_and_persistence, filtered_max_points_and_persistence) = (
        diversify_extremum_points_and_persistence(filtered_extremum_points_and_persistence, level_set="upper")
    )

    # Sorting maximum points (and, as a consequence, the corresponding minimum points) w.r.t. persistence:
    argsorting = np.argsort(list(zip(*filtered_max_points_and_persistence))[1]).tolist()

    if len(filtered_min_points_and_persistence) == 0:
        filtered_min_points = []
        persistence_of_filtered_min_points = []
    else:
        filtered_min_points = np.array(list(zip(*filtered_min_points_and_persistence))[0])[argsorting[:-1]].tolist()
        persistence_of_filtered_min_points = np.array(list(zip(*filtered_min_points_and_persistence))[1])[
            argsorting[:-1]
        ].tolist()

    # Indices of maximum points and their persistence:
    filtered_max_points = np.array(list(zip(*filtered_max_points_and_persistence))[0])[argsorting].tolist()
    persistence_of_filtered_max_points = np.array(list(zip(*filtered_max_points_and_persistence))[1])[
        argsorting
    ].tolist()

    return (
        filtered_min_points,
        persistence_of_filtered_min_points,
        filtered_max_points,
        persistence_of_filtered_max_points,
    )



# Copyright (C) 2024 Roberto Rossini <roberroso@uio.no>
#
# SPDX-License-Identifier: MIT

import pathlib

import h5py
import pytest

from stripepy import main

testdir = pathlib.Path(__file__).resolve().parent.parent


@pytest.mark.end2end
class TestStripePyCall:
    @staticmethod
    def setup_class():
        test_files = [
            testdir / "data" / "4DNFI9GMP2J8.mcool",
        ]

        for f in test_files:
            if not f.exists():
                raise RuntimeError(
                    f'unable to find file "{f}". Did you download the test files prior to running pytest?'
                )

    @staticmethod
    def test_stripepy_call(tmpdir):
        testfile = testdir / "data" / "4DNFI9GMP2J8.mcool"
        resolution = 10_000

        args = [
            "call",
            str(testfile),
            str(resolution),
            "--glob-pers-min",
            "0.10",
            "--loc-pers-min",
            "0.33",
            "--loc-trend-min",
            "0.25",
            "--output-folder",
            str(tmpdir),
        ]
        main(args)

        outfile = pathlib.Path(tmpdir) / testfile.stem / str(resolution) / "results.hdf5"

        assert outfile.is_file()
        assert h5py.File(outfile).attrs.get("format", "unknown") == "HDF5::StripePy"



# Copyright (C) 2024 Roberto Rossini <roberros@uio.no>
#
# SPDX-License-Identifier: MIT

from typing import Sequence, Tuple

import numpy as np
from numpy.typing import NDArray


def sort_based_on_arg0(*vectors: Sequence) -> Tuple[NDArray]:
    """
    Sort two or more sequences of objects based on the first sequence of objects.

    Parameters
    ----------
    vectors: two or more sequences to be sorted

    Returns
    -------
    Tuple[NDArray]
        the sorted sequences as numpy.ndarray
    """
    if len(vectors) < 2:
        raise ValueError("please specify at least two sequences")

    for v in vectors[1:]:
        assert len(vectors[0]) == len(v)

    if len(vectors[0]) == 0:
        return tuple((np.array(v) for v in vectors))  # noqa

    permutation = np.argsort(vectors[0])

    return tuple((np.array(v)[permutation] for v in vectors))  # noqa



# Copyright (C) 2024 Andrea Raffo <andrea.raffo@ibv.uio.no>
#
# SPDX-License-Identifier: MIT

from functools import partial
from multiprocessing import Pool

import matplotlib.pyplot as plt
import numpy as np

from . import TDA
from .regressions import _compute_wQISA_predictions


def find_horizontal_domain(pd, coarse_h_domain, max_width=1e9):

    # Unpacking:
    MP, L_mP, R_mP = coarse_h_domain

    # Left and sides of candidate:
    L_interval = np.flip(pd[L_mP : MP + 1])
    R_interval = pd[MP : R_mP + 1]

    # LEFT INTERVAL
    L_interval_shifted = np.append(L_interval[1:], [max(pd) + 1], axis=0)
    L_bound = np.where(L_interval - L_interval_shifted < 0)[0][0] + 1
    # L_interval_restr = L_interval[:L_bound]
    # L_interval_shifted_restr = L_interval_shifted[:L_bound]
    # L_bound = np.argmax(L_interval_restr - L_interval_shifted_restr) + 1
    L_bound = min(L_bound, max_width)

    # RIGHT INTERVAL
    R_interval_shifted = np.append(R_interval[1:], [max(pd) + 1], axis=0)
    R_bound = np.where(R_interval - R_interval_shifted < 0)[0][0] + 1
    # R_interval_restr = R_interval[:R_bound]
    # R_interval_shifted_restr = R_interval_shifted[:R_bound]
    # R_bound = np.argmax(R_interval_restr - R_interval_shifted_restr) + 1
    R_bound = min(R_bound, max_width)

    return [max(MP - L_bound, 0), min(MP + R_bound, len(pd))]


def find_lower_v_domain(I, VIoIs2plot, threshold_cut, max_height, min_persistence, output_folder, it):

    # For each slice (hosting a seed site):
    n, seed_site, HIoI = it

    # Standardized Local 1D pseudo-distribution for current HIoI:
    rows = slice(seed_site, min(seed_site + max_height, I.shape[0]))
    cols = slice(HIoI[0], HIoI[1])
    I_nb = I[rows, :].tocsc()[:, cols].toarray()

    Y = np.sum(I_nb, axis=1)
    Y /= max(Y)

    # Lower boundary:
    X_tr = np.array(range(seed_site, seed_site + len(Y)))
    Y_hat = _compute_wQISA_predictions(Y, 5)  # Basically: average of a 2-"pixel" neighborhood

    # Peaks:
    if min_persistence is None:

        candida_bound = np.where(np.array(Y_hat) < threshold_cut)[0]
        if len(candida_bound) == 0:
            candida_bound = [len(Y_hat) - 1]

        # Vertical domain + no peak:
        Vdomain_and_peaks = ([seed_site, seed_site + candida_bound[0]], None)

    else:
        _, _, loc_Maxima, loc_pers_of_Maxima = TDA.TDA(Y, min_persistence=min_persistence)
        candida_bound = [max(loc_Maxima)]

        # Consider as min_persistence is set to None:
        if len(loc_Maxima) < 2:
            candida_bound = np.where(np.array(Y_hat) < threshold_cut)[0]
            if len(candida_bound) == 0:
                candida_bound = [len(Y_hat) - 1]

        # Vertical domain + peaks:
        Vdomain_and_peaks = ([seed_site, seed_site + candida_bound[0]], list(np.array(loc_Maxima[:-1]) + seed_site))

    # In case some VIoIs are meant to be plotted:
    if VIoIs2plot is not None:
        if n in VIoIs2plot:
            fig, ax = plt.subplots(1, 1)
            ax.plot(X_tr, Y, color="red", linewidth=0.5, linestyle="solid")
            ax.plot(X_tr, Y_hat, color="black", linewidth=0.5, linestyle="solid")
            if min_persistence is not None:
                ax.plot(
                    [seed_site + a for a in loc_Maxima[:-1]],
                    np.array(Y)[loc_Maxima[:-1]].tolist(),
                    color="blue",
                    marker=".",
                    linestyle="",
                    markersize=8 * 1.5,
                )
            ax.plot(
                [seed_site + candida_bound[0], seed_site + candida_bound[0]],
                [0.0, 1.0],
                color="blue",
                linewidth=1.0,
                linestyle="dashed",
            )
            plt.xlim((X_tr[0], X_tr[-1]))
            plt.ylim((0, 1))
            fig.set_dpi(256)
            fig.tight_layout()
            plt.savefig(f"{output_folder}/LT_local-pseudo-distrib_{seed_site}.jpg")
            plt.close()
    return Vdomain_and_peaks


def find_upper_v_domain(I, VIoIs2plot, threshold_cut, max_height, min_persistence, output_folder, it):

    # For each slice (hosting a seed site):
    n, seed_site, HIoI = it

    # Standardized Local 1D pseudo-distribution for current HIoI:
    rows = slice(max(seed_site - max_height, 0), seed_site)
    cols = slice(HIoI[0], HIoI[1])
    I_nb = I[rows, :].tocsc()[:, cols].toarray()
    Y = np.flip(np.sum(I_nb, axis=1))
    Y /= max(Y)

    # Upper boundary:
    X_tr = np.array(range(seed_site, seed_site + len(Y)))
    # Y_hat = compute_predictions(X, Y, 10.)
    Y_hat = _compute_wQISA_predictions(Y, 5)

    # Peaks:
    if min_persistence is None:

        candida_bound = np.where(np.array(Y_hat) < threshold_cut)[0]
        if len(candida_bound) == 0:
            candida_bound = [len(Y_hat) - 1]

        # Vertical domain + no peak:
        Vdomain_and_peaks = ([seed_site - candida_bound[0], seed_site], None)

    else:
        _, _, loc_Maxima, loc_pers_of_Maxima = TDA.TDA(Y, min_persistence=min_persistence)
        candida_bound = [max(loc_Maxima)]

        # Consider as min_persistence is set to None:
        if len(loc_Maxima) < 2:
            candida_bound = np.where(np.array(Y_hat) < threshold_cut)[0]
            if len(candida_bound) == 0:
                candida_bound = [len(Y_hat) - 1]

        # Vertical domain + peaks:
        Vdomain_and_peaks = ([seed_site - candida_bound[0], seed_site], list(seed_site - np.array(loc_Maxima[:-1])))

    # In case some VIoIs are meant to be plotted:
    if VIoIs2plot is not None:
        if n in VIoIs2plot:
            fig, ax = plt.subplots(1, 1)
            ax.plot(X_tr, Y, color="red", linewidth=0.5, linestyle="solid")
            ax.plot(X_tr, Y_hat, color="black", linewidth=0.5, linestyle="solid")
            if min_persistence is not None:
                ax.plot(
                    [seed_site + a for a in loc_Maxima[:-1]],
                    np.array(Y)[loc_Maxima[:-1]].tolist(),
                    color="blue",
                    marker=".",
                    linestyle="",
                    markersize=8 * 1.5,
                )
            ax.plot(
                [seed_site + candida_bound[0], seed_site + candida_bound[0]],
                [0.0, 1.0],
                color="blue",
                linewidth=1.0,
                linestyle="dashed",
            )
            plt.xlim((X_tr[0], X_tr[-1]))
            plt.ylim((0, 1))
            fig.set_dpi(256)
            fig.tight_layout()
            plt.savefig(f"{output_folder}/UT_local-pseudo-distrib_{seed_site}.jpg")
            plt.close()

    return Vdomain_and_peaks


def find_HIoIs(pd, seed_sites, seed_site_bounds, max_width, map=map):
    """
    :param pd:                  acronym for pseudo-distribution, but can be any 1D array representing a uniformly-sample
                                scalar function works
    :param seed_sites:          maximum values in the pseudo-distribution (i.e., genomic coordinates hosting linear
                                patterns)
    :param seed_site_bounds:    for the i-th entry of seed_sites:
                                (*) seed_site_bounds[i] is the left boundary
                                (*) seed_site_bounds[i+1] is the right boundary
    :param max_width:           maximum width allowed
    :param map:                 alternative implementation of the built-in map function. Can be used to e.g. run this step in parallel by passing multiprocessing.Pool().map.
    :return:
    HIoIs                       list of lists, where each sublist is a pair consisting of the left and right boundaries
    """
    assert len(seed_site_bounds) == len(seed_sites) + 1

    iterable_input = [
        (seed_site, seed_site_bounds[num_MP], seed_site_bounds[num_MP + 1])
        for num_MP, seed_site in enumerate(seed_sites)
    ]

    HIoIs = list(map(partial(find_horizontal_domain, pd, max_width=max_width), iterable_input))

    # Handle possible overlapping intervals:
    for i in range(len(HIoIs) - 1):
        current_pair = HIoIs[i]
        next_pair = HIoIs[i + 1]

        if current_pair[1] > next_pair[0]:  # Check for intersection
            next_pair[0] = current_pair[1]  # Modify the second pair

    return HIoIs


def find_VIoIs(
    I,
    seed_sites,
    HIoIs,
    max_height,
    threshold_cut=0.1,
    min_persistence=0.20,
    VIoIs2plot=None,
    output_folder=None,
    where="lower",
    map=map,
):

    # Triplets to use in multiprocessing:
    iterable_input = [(n, seed_site, HIoI) for n, (seed_site, HIoI) in enumerate(zip(seed_sites, HIoIs))]

    # Lower-triangular part of the Hi-C matrix:
    if where == "lower":
        Vdomains_and_peaks = map(
            partial(find_lower_v_domain, I, VIoIs2plot, threshold_cut, max_height, min_persistence, output_folder),
            iterable_input,
        )

        VIoIs, peak_locs = list(zip(*Vdomains_and_peaks))

    # Upper-triangular part of the Hi-C matrix:
    elif where == "upper":
        # HIoIs = pool.map(partial(find_h_domain, pd), iterable_input)
        Vdomains_and_peaks = map(
            partial(find_upper_v_domain, I, VIoIs2plot, threshold_cut, max_height, min_persistence, output_folder),
            iterable_input,
        )

        VIoIs, peak_locs = list(zip(*Vdomains_and_peaks))

    return VIoIs, peak_locs



# Copyright (C) 2024 Roberto Rossini <roberroso@uio.no>
#
# SPDX-License-Identifier: MIT

import json
import math
import pathlib
import shutil
import tempfile

import hictkpy as htk
import numpy as np
import pytest
from scipy.sparse import csr_matrix

from stripepy.IO import (
    Result,
    ResultFile,
    create_folders_for_plots,
    remove_and_create_folder,
)
from stripepy.utils.stripe import Stripe

from .common.cool import generate_singleres_test_file

testdir = pathlib.Path(__file__).resolve().parent.parent


def _directory_is_empty(path) -> bool:
    path = pathlib.Path(path)
    assert path.is_dir()
    return next(path.iterdir(), None) is None  # noqa


@pytest.mark.unit
def test_folders_for_plots(tmpdir):
    with tempfile.TemporaryDirectory(dir=tmpdir) as tmpdir:
        test_paths = [
            pathlib.Path(tmpdir),
            pathlib.Path(tmpdir) / "dir2",
        ]

        for test_dir in test_paths:
            if test_dir.exists():
                shutil.rmtree(test_dir)

            assert create_folders_for_plots(test_dir) == [
                test_dir,
                test_dir / "1_preprocessing",
                test_dir / "2_TDA",
                test_dir / "3_shape_analysis",
                test_dir / "4_biological_analysis",
                test_dir / "3_shape_analysis" / "local_pseudodistributions",
            ]


@pytest.mark.unit
class TestRemoveAndCreateFolder:
    # RuntimeError(f"output folder {path} already exists. Pass --force to overwrite it.")
    @staticmethod
    def test_create_new_folder(tmpdir):
        with tempfile.TemporaryDirectory(dir=tmpdir) as tmpdir:
            tmpdir = pathlib.Path(tmpdir)

            test_dir = tmpdir / "out"
            remove_and_create_folder(test_dir, force=True)
            assert test_dir.is_dir()
            assert _directory_is_empty(test_dir)

            with pytest.raises(RuntimeError, match="already exists. Pass --force to overwrite it"):
                remove_and_create_folder(test_dir, force=False)

    @staticmethod
    def test_overwrite_existing_folder(tmpdir):
        with tempfile.TemporaryDirectory(dir=tmpdir) as tmpdir:
            tmpdir = pathlib.Path(tmpdir)

            test_dir = tmpdir / "out"

            (test_dir / "dir").mkdir(parents=True)
            (test_dir / "file.txt").touch()

            assert not _directory_is_empty(test_dir)
            remove_and_create_folder(test_dir, force=True)
            assert _directory_is_empty(test_dir)


@pytest.mark.unit
def test_create_folders_for_plots(tmpdir):
    with tempfile.TemporaryDirectory(dir=tmpdir) as tmpdir:
        tmpdir = pathlib.Path(tmpdir)

        test_dir = tmpdir / "out"
        result = create_folders_for_plots(test_dir)
        assert isinstance(result, list)
        assert test_dir.is_dir()


@pytest.mark.unit
class TestResult:
    def test_ctor(self):
        res = Result("chr1")

        assert res.empty
        assert res.chrom == "chr1"
        assert res.roi is None

    def test_setters(self):
        res = Result("chr1")

        res.set_min_persistence(1.0)
        with pytest.raises(RuntimeError, match="has already been set"):
            res.set_min_persistence(1.0)

        res.set_roi({"foobar": [1, 2, 3]})
        with pytest.raises(RuntimeError, match="has already been set"):
            res.set_roi({})

        assert res.roi is not None
        assert "foobar" in res.roi
        assert res.roi["foobar"] == [1, 2, 3]

        with pytest.raises(AttributeError, match='No attribute named "invalid name"'):
            res.set("invalid name", [], "LT")

        with pytest.raises(ValueError, match="Location should be UT or LT"):
            res.set("all_minimum_points", [], "invalid location")

        res.set("all_minimum_points", [1, 2, 3], "LT")
        res.set("all_minimum_points", [4, 5, 6], "UT")

        with pytest.raises(RuntimeError, match="has already been set"):
            res.set("all_minimum_points", [], "LT")

        assert (res.get("all_minimum_points", "LT") == [1, 2, 3]).all()
        assert (res.get("all_minimum_points", "UT") == [4, 5, 6]).all()

    def test_getters(self):
        res = Result("chr1")

        res.set("all_minimum_points", [1, 2, 3], "LT")

        with pytest.raises(RuntimeError, match="Attribute .* is not set"):
            res.min_persistence  # noqa

        res.set_min_persistence(1.23)
        assert res.min_persistence == 1.23

        with pytest.raises(AttributeError, match='No attribute named "invalid name"'):
            res.get("invalid name", "LT")

        with pytest.raises(ValueError, match="Location should be UT or LT"):
            res.get("all_minimum_points", "invalid location")

        with pytest.raises(RuntimeError, match='Attribute "all_minimum_points" for "UT" is not set'):
            res.get("all_minimum_points", "UT")

        with pytest.raises(RuntimeError, match='Attribute "all_maximum_points" for "LT" is not set'):
            res.get("all_maximum_points", "LT")

        assert len(res.get("stripes", "LT")) == 0

        assert (res.get("all_minimum_points", "LT") == [1, 2, 3]).all()

    def test_stripe_getters(self):
        res = Result("chr1")

        stripes = [Stripe(10, 1.23)]
        stripes[0].set_vertical_bounds(5, 10)
        stripes[0].set_horizontal_bounds(8, 12)
        stripes[0].compute_biodescriptors(csr_matrix((15, 15), dtype=float))

        with pytest.raises(AttributeError, match='does not have an attribute named "foobar"'):
            res.get_stripes_descriptor("foobar", "UT")

        with pytest.raises(ValueError, match="Location should be UT or LT"):
            res.get_stripes_descriptor("seed", "foobar")

        res.set("stripes", stripes, "LT")

        assert len(res.get_stripe_geo_descriptors("UT")) == 0
        assert len(res.get_stripe_bio_descriptors("UT")) == 0

        df = res.get_stripe_geo_descriptors("LT")
        assert df.columns.tolist() == [
            "seed",
            "top_persistence",
            "left_bound",
            "right_bound",
            "top_bound",
            "bottom_bound",
        ]
        assert len(df) == 1

        assert df["seed"].iloc[0] == 10
        assert df["top_persistence"].iloc[0] == 1.23
        assert df["left_bound"].iloc[0] == 8
        assert df["right_bound"].iloc[0] == 12
        assert df["top_bound"].iloc[0] == 5
        assert df["bottom_bound"].iloc[0] == 10

        df = res.get_stripe_bio_descriptors("LT")
        assert df.columns.tolist() == ["inner_mean", "outer_mean", "rel_change", "inner_std"]
        assert len(df) == 1

        assert df["inner_mean"].iloc[0] == 0
        assert df["outer_mean"].iloc[0] == 0
        assert df["rel_change"].iloc[0] == -1
        assert df["inner_std"].iloc[0] == 0


@pytest.mark.unit
class TestResultFile:
    def test_ctor(self, tmpdir):
        tmpdir = pathlib.Path(tmpdir)

        path = testdir / "data" / "results_4DNFI9GMP2J8_v1.hdf5"
        with ResultFile(path) as f:
            assert f.path == path

        with pytest.raises(OSError):
            ResultFile(tmpdir / "asdf123.hdf5")

        path = pathlib.Path(tmpdir) / "test.hdf5"
        f = ResultFile(path, "w")
        assert f.path == path

        with pytest.raises(ValueError):
            ResultFile(path, "a")

    def test_properties(self):
        path = testdir / "data" / "results_4DNFI9GMP2J8_v1.hdf5"
        with ResultFile(path) as f:
            assert f.path == path
            assert f.assembly == "unknown"
            assert f.resolution == 10_000

            try:
                f.creation_date  # noqa
            except ValueError:
                pytest.fail("creation-date attribute is not a valid date")

            assert f.format == "HDF5::StripePy"
            assert f.format_url == "https://github.com/paulsengroup/StripePy"
            assert f.format_version == 1
            assert f.generated_by.startswith("StripePy")

            try:
                f.metadata  # noqa
            except json.decoder.JSONDecodeError:
                pytest.fail("metadata attribute is not a valid JSON")

            assert f.normalization is None

            assert len(f.chromosomes) == 24
            assert "chr1" in f.chromosomes

    def test_getters(self):
        path = testdir / "data" / "results_4DNFI9GMP2J8_v1.hdf5"
        with ResultFile(path) as f:
            assert math.isclose(f.get_min_persistence("chr1"), 0.05)
            with pytest.raises(KeyError):
                f.get_min_persistence("foobar")

            df = f.get("chr1", "pseudodistribution", "LT")

            assert "pseudodistribution" in df
            assert len(df) == 24896

            df = f.get("chr1", "geo_descriptors", "LT")

            geo_descriptor_cols = ["seed", "top_persistence", "left_bound", "right_bound", "top_bound", "bottom_bound"]
            assert len(df.columns) == len(geo_descriptor_cols)
            assert (df.columns == geo_descriptor_cols).all()
            assert len(df) == 1305

            df = f.get("chr1", "bio_descriptors", "LT")

            bio_descriptor_cols = ["inner_mean", "outer_mean", "rel_change", "inner_std"]
            assert len(df.columns) == len(bio_descriptor_cols)
            assert (df.columns == bio_descriptor_cols).all()
            assert len(df) == 1305

            with pytest.raises(KeyError):
                f.get("foobar", "pseudodistribution", "LT")

            with pytest.raises(KeyError):
                f.get("chr1", "foobar", "LT")

            with pytest.raises(ValueError):
                f.get("chr1", "pseudodistribution", "foobar")

    def test_file_creation(self, tmpdir):
        tmpdir = pathlib.Path(tmpdir)

        clr_file = generate_singleres_test_file(tmpdir / "test.cool", 10_000)
        chroms = htk.File(clr_file).chromosomes()
        chrom = tuple(chroms.keys())[0]

        points = [1, 2, 3]
        persistence = [4.0, 5.0, 6.0]
        pseudoditribution = [7.0, 8.0, 9.0]

        # Create a mock ResultFile
        with ResultFile(tmpdir / "results.hdf5", "w") as f:
            f.init_file(htk.File(clr_file), "weight", {"key": "value"})

            res = Result(chrom)
            res.set_min_persistence(1.23)

            for location in ["UT", "LT"]:
                res.set("pseudodistribution", pseudoditribution, location)

                for key in [
                    "all_minimum_points",
                    "all_maximum_points",
                    "persistent_minimum_points",
                    "persistent_maximum_points",
                ]:
                    res.set(key, points, location)

                for key in [
                    "persistence_of_all_minimum_points",
                    "persistence_of_all_maximum_points",
                    "persistence_of_minimum_points",
                    "persistence_of_maximum_points",
                ]:
                    res.set(key, persistence, location)

            f.write_descriptors(res)

        with ResultFile(tmpdir / "results.hdf5") as f:
            assert f.chromosomes == chroms
            assert f.normalization == "weight"
            assert f.metadata.get("key", "missing") == "value"

            for location in ["UT", "LT"]:
                assert np.allclose(
                    f.get(chrom, "pseudodistribution", location)["pseudodistribution"], pseudoditribution
                )

                for key in ["all_minimum_points", "all_maximum_points"]:
                    assert (f.get(chrom, key, location)[key] == points).all()

                for key in ["persistence_of_all_minimum_points", "persistence_of_all_maximum_points"]:
                    assert np.allclose(f.get(chrom, key, location)[key], persistence)



#!/usr/bin/env python3

import argparse
import pathlib
import subprocess as sp
import time


# Create a custom formatter to allow multiline and bulleted descriptions
class CustomFormatter(argparse.RawTextHelpFormatter):
    def _fill_text(self, text, width, indent):
        return "".join([indent + line + "\n" for line in text.splitlines()])


def _input_dir_checked(arg: str) -> pathlib.Path:
    input_dir = pathlib.Path(arg)
    if input_dir.exists() and input_dir.is_dir():
        return pathlib.Path(arg)

    raise FileNotFoundError(f'Input folder "{arg}" is not reachable: folder does not exist')


def _output_dir_checked(arg: str) -> pathlib.Path:
    parent = pathlib.Path(arg).parent
    if parent.exists() and parent.is_dir():
        return pathlib.Path(arg)

    raise FileNotFoundError(f'Output folder "{arg}" is not reachable: parent folder does not exist')


def _probability(arg) -> float:
    if 0 <= (n := float(arg)) <= 1:
        return n

    raise ValueError("Not a valid probability")


def make_cli():
    cli = argparse.ArgumentParser(
        description="This script runs StripePy over StripeBench, a benchmark containing 64 simulated Hi-C contact maps generated "
        "via the computational tool MoDLE at different resolutions, contact densities and noise levels.",
        formatter_class=CustomFormatter,
    )

    cli.add_argument(
        "stripepy-exec",
        type=pathlib.Path,
        help="Path to StripePy executable",
    )

    cli.add_argument(
        "stripebench-path",
        type=_input_dir_checked,
        help="Path to the StripeBench dataset, which can be downloaded from XX.",
        # TODO complete when benchmark is online
    )

    cli.add_argument(
        "-b",
        "--genomic-belt",
        type=int,
        default=5_000_000,
        help="Radius of the band, centred around the diagonal, where the search is restricted to "
        "(in bp). The value used for the StripeBench benchmark is here set as default.",
    )

    cli.add_argument(
        "-o",
        "--output-folder",
        type=_output_dir_checked,
        default=pathlib.Path("."),
        help="Path to the folder where the user wants the output to be placed (default: current folder).",
    )

    cli.add_argument(
        "--max-width",
        type=int,
        default=20_000,
        help="Maximum stripe width, in bp.",
    )

    cli.add_argument(
        "--glob-pers-min",
        type=_probability,
        default=0.05,
        help="Threshold value between 0 and 1 to filter persistence maxima points and identify loci of interest, "
        "aka seeds. The value used for the StripeBench benchmark is here set as default.",
    )

    cli.add_argument(
        "--loc-pers-min",
        type=_probability,
        default=0.25,
        help="Threshold value between 0 and 1 to find peaks in signal in a horizontal domain while estimating the "
        "height of a stripe. The value used for the StripeBench benchmark is here set as default.",
    )

    cli.add_argument(
        "--loc-trend-min",
        type=_probability,
        default=0.05,
        help="Threshold value between 0 and 1 to estimate the height of a stripe; the higher this value, the shorter "
        "the stripe; it is always used when --constrain-heights is set to 'False', but could be necessary also "
        "when --constrain-heights is 'True' and no persistent maximum other than the global maximum is found. "
        "The value used for the StripeBench benchmark is here set as default.",
    )

    cli.add_argument(
        "-f",
        "--force",
        action="store_true",
        default=False,
        help="Overwrite existing file(s).",
    )

    return cli


def run_stripepy(
    stripepy_exec,
    path_to_mcool,
    resolution,
    genomic_belt,
    output_folder,
    max_width,
    glob_pers_min,
    loc_pers_min,
    loc_trend_min,
    force,
):
    args = [
        stripepy_exec,
        "call",
        path_to_mcool,
        resolution,
        "-b",
        str(genomic_belt),
        "-o",
        output_folder,
        "--max-width",
        str(max_width),
        "--glob-pers-min",
        str(glob_pers_min),
        "--loc-pers-min",
        str(loc_pers_min),
        "--loc-trend-min",
        str(loc_trend_min),
    ]

    if force:
        args.append("--force")

    sp.check_call(args)


def main():
    args = vars(make_cli().parse_args())

    with open(f"{args["output_folder"]}/output.log", "w") as f:
        t0 = time.time()
        resolutions = ["5000", "10000", "25000", "50000"]
        contact_densities = ["1", "5", "10", "15"]
        noise_levels = ["0", "5000", "10000", "15000"]
        for contact_density in contact_densities:
            for noise_level in noise_levels:
                this_contact_map = (
                    args["stripebench-path"]
                    / "data"
                    / f"grch38_h1_rad21_{contact_density}_{noise_level}"
                    / f"grch38_h1_rad21_{contact_density}_{noise_level}.mcool"
                )
                for resolution in resolutions:
                    run_stripepy(
                        args["stripepy-exec"],
                        this_contact_map,
                        resolution,
                        args["genomic_belt"],
                        args["output_folder"],
                        args["max_width"],
                        args["glob_pers_min"],
                        args["loc_pers_min"],
                        args["loc_trend_min"],
                        args["force"],
                    )
        delta = time.time() - t0
        print("Total time: ", file=f)
        print(delta, file=f)


if __name__ == "__main__":
    main()
